;; Generated package description from geiser.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "geiser" "0.33.1.0.20260523.150214" "GNU Emacs and Scheme talk to each other" '((emacs "27.1") (project "0.8.1")) :commit "84c25e9683a18d00387b6c16b0cee66269536c3c" :authors '(("Jose Antonio Ortega Ruiz" . "(jao@gnu.org)")) :maintainer '("Jose Antonio Ortega Ruiz" . "(jao@gnu.org)") :keywords '("languages" "scheme" "geiser") :url "https://gitlab.com/emacs-geiser/")
