html {
    padding:0px;margin:0px;
    background-color:white;
    color:black;
    font-family:sans-serif;
    line-height:160%
}

body {
    padding:10px 10% 10px 10%;
    margin:0px;
    text-align:justify;
    /* width: 750px; */
}

a { color:black; weight=normal }
a:hover { color: #a22 }

hr { height:0; color:white }

h2 {
    font-size:120%;
    text-transform:uppercase
}

h3 {
    font-size:100%;
    text-transform:uppercase
}

pre.code {
    display:block;
    padding:0px;
    margin-bottom:0px
}

code {
    background-color:rgb(250, 250, 250);
    border:1px solid rgb(200, 200, 200);
    padding-left:4px;
    padding-right:4px
}

pre.example {
    background-color:rgb(250, 250, 250);
    border:1px solid rgb(200, 200, 200);
    padding-left:4px;
    padding-right:4px;
}

table { width: 100%; }

img {
    display:block;
    margin:10px auto 10px auto;
    border:none
}

img.floatleft {
    float: left;
    margin: 4px;
    padding-right: 1em;
}

img.floatright {
    float: right;
    margin: 4px;
    padding-left: 1em;
}

ul {
    list-style-type:square;
    padding-left:1em;
    margin-left:1em
}

kbd {
    font-weight: bold;
}

div.navigation {
    background-color: #efebe7;
    line-height: 100%;
    top-margin: 5px;
    padding: 3px;
    /* background-color: #000000; */
    /* border-top-style: solid; */
    /* border-bottom-style: solid; */
    /* border-width: 0px; */
    /* border-top-width: 4px; */
    /* border-color: rgb(200, 200, 200); */
}

div.version {
    /* background-color: #efebe7; */
    line-height: 100%;
    padding: 3px;
    margin-right: 3px;
    font-size: smaller;
    float: right;
    font-style: italic;
}

div.version a {
    text-decoration: none;
}

.navlink {
    text-decoration: none;
    /* color: white; */
}

.index-cp {
    line-height:120%;
    font-size: smaller;
}
